module chain_analysis_mod
use types_mod
use maths_mod
use global_vars_mod, only : one_sigma, two_sigma, atmos, n_param_per_component, atmos, Params, Stokes_Syn, Line, Region, Hydro, Error, obs
use multinest_mod, only : synthesize_profiles
implicit none

contains

!------------------------------------------------------------
! Read the chains
!------------------------------------------------------------
	subroutine read_chains_multinest(fich_chain, chain_analysis)
	type(markov_chain_type) :: chain_analysis
	character(len=120) :: fich_chain
	integer :: nparams, chain_length, i, j
	logical :: exists

		i = 0
		
		inquire(file=trim(adjustl(fich_chain))//'post_equal_weights.dat', exist=exists)
		if (.not.exists) then
			write(*,*) 'File not found : ', trim(adjustl(fich_chain))//'post_equal_weights.dat'
			stop
		endif
		
		open(unit=12,file=trim(adjustl(fich_chain))//'post_equal_weights.dat',action='read',status='old')
		do while (.not.eof(12))
			read(12,*)
			i = i + 1
		enddo
		close(12)
		
		chain_analysis%nlength = i
		print *, 'Length of posterior samples : ', chain_analysis%nlength
		print *, 'Number of parameters : ', chain_analysis%nparam
		
		allocate(chain_analysis%chain(chain_analysis%nparam+1,chain_analysis%nlength))
		open(unit=12,file=trim(adjustl(fich_chain))//'post_equal_weights.dat',action='read',&
			status='old')
		do i = 1, chain_analysis%nlength
			read(12,*) (chain_analysis%chain(j,i),j=1,chain_analysis%nparam+1)
		enddo
		close(12)

! Read evidence
		inquire(file=trim(adjustl(fich_chain))//'stats.dat', exist=exists)
		if (.not.exists) then
			write(*,*) 'File not found : ', trim(adjustl(fich_chain))//'stats.dat'
			stop
		endif
		
		open(unit=12,file=trim(adjustl(fich_chain))//'stats.dat',action='read',status='old')
		read(12,FMT='(16X,E20.12)') chain_analysis%evidence
		close(12)
		
	end subroutine read_chains_multinest
	
!------------------------------------------------------------
! Remove burn-in
!------------------------------------------------------------
	subroutine remove_burnin(chain_analysis)
	type(markov_chain_type) :: chain_analysis
	real(kind=8), pointer :: chain_new(:,:), temp(:)
	real(kind=8) :: percent
	integer :: burn_in
				
		write(*,*) 'Burn-in of ', chain_analysis%burnin, '%'
		
		burn_in = chain_analysis%niter_max * chain_analysis%burnin / 100.d0
		if (burn_in > chain_analysis%niter_max) burn_in = chain_analysis%niter_max
						
		write(*,*) 'Leaving a chain of length : ', chain_analysis%niter_max-burn_in

! Change length of chain without burn-in
		chain_analysis%niter_max = chain_analysis%niter_max-burn_in

! Re-allocate new memory
		allocate(chain_new(chain_analysis%nparam,chain_analysis%niter_max))
		chain_new = chain_analysis%chain(:,burn_in:)
		
		deallocate(chain_analysis%chain)
		allocate(chain_analysis%chain(chain_analysis%nparam,chain_analysis%niter_max))
		chain_analysis%chain = chain_new
				
		deallocate(chain_new)

! Do the same with the acceptance rate and the log-likelihood
		allocate(temp(chain_analysis%niter_max))
		temp = chain_analysis%posterior(burn_in:)		
		deallocate(chain_analysis%posterior)
		allocate(chain_analysis%posterior(chain_analysis%niter_max))
		chain_analysis%posterior = temp

		temp = chain_analysis%acceptance_rate(burn_in:)
		deallocate(chain_analysis%acceptance_rate)
		allocate(chain_analysis%acceptance_rate(chain_analysis%niter_max))
		chain_analysis%acceptance_rate = temp
				
		deallocate(temp)
			
	end subroutine remove_burnin

!------------------------------------------------------------
! Get the optimal binning for building a histogram of a variable
! Ref: Freedman, D, & Diaconis, P. (1981). "On the histogram as a density
!      estimator: L2 theory". Zeitschrift für Wahrscheinlichkeitstheorie und
!      verwandte Gebiete 57 (4): 453–476
!  BIN=2*IQR(var)/n^(1/3)   with n the number of elements of the array 'var'
!------------------------------------------------------------
	function optbin(var)
	real(kind=8) :: optbin, var(:)
	integer :: n, quart1, quart3
	integer, allocatable :: ind(:)
	
		n = size(var)
		allocate(ind(n))
		
		call qsortd(var,ind,n)
		
		quart1 = n*0.25
		quart3 = n*0.75
		
		optbin = 2.0*(var(ind(quart3)) - var(ind(quart1))) / (n*1.0)**(1.0/3.0)

		deallocate(ind)
		
	end function optbin

!------------------------------------------------------------
! Estimate 0.2%, 2.5%, 16%, 50%, 84%, 97.5%, 99.8% percentiles
! of a given chain
!------------------------------------------------------------
	function percentile_summary(var)
	real(kind=8) :: percentile_summary(7), var(:)
	integer :: n, loc
	integer, allocatable :: ind(:)
	
		n = size(var)
		allocate(ind(n))
		
		call qsortd(var,ind,n)

		loc = n * 0.2 / 100.0
		percentile_summary(1) = var(ind(loc))
		loc = n * 2.5 / 100.0
		percentile_summary(2) = var(ind(loc))
		loc = n * 16.0 / 100.0
		percentile_summary(3) = var(ind(loc))
		loc = n * 50.0 / 100.0
		percentile_summary(4) = var(ind(loc))
		loc = n * 84.0 / 100.0
		percentile_summary(5) = var(ind(loc))
		loc = n * 97.5 / 100.0
		percentile_summary(6) = var(ind(loc))
		loc = n * 99.8 / 100.0
		percentile_summary(7) = var(ind(loc))

		deallocate(ind)
		
	end function percentile_summary

!------------------------------------------------------------
! 1D histograms
!------------------------------------------------------------
	subroutine oned_histogram(chain, x, yGauss, yStep)
	real(kind=8) :: chain(:)
	real(kind=8) :: step, xmin, xmax
	real(kind=8), pointer :: x(:), yGauss(:), yStep(:)
	real(kind=8), allocatable :: xx(:)
	integer :: i, n, nn, ncut
	real(kind=8) :: wei, norm, error_norm, sig
	
! Bins
		step = optbin(chain)
				
		xmin = minval(chain)
		xmax = maxval(chain)
		n = (xmax-xmin) / step		
		
		step = (xmax-xmin) / float(n)
		
! Variables
		allocate(x(n))
		allocate(yGauss(n))
		allocate(yStep(n))
		
		do i = 1, n
			x(i) = xmin + step * (i-1.0)
		enddo
		
! Variables for the normalization
		if (n <= 50) then
			nn = 50
			allocate(xx(nn))
			do i = 1, nn
				xx(i) = xmin + (xmax-xmin)/nn * i
			enddo
		else
			allocate(xx(n))
			xx = x
		endif
				
! Doing the histogram
		sig = 1.2d0
		do i = 1, n
			ncut = count(chain >= x(i)-step/2.d0 .and. chain < x(i)+step/2.0)
			
! Smoothing kernel
			wei = sum(exp(-0.50 * (chain-x(i))**2 / (sig*step)**2.0))
			norm = int_tabulated(xx, exp(-0.50 * (xx-x(i))**2 / (sig*step)**2.0))
			yGauss(i) = wei / norm
			yStep(i) = ncut
		enddo
		
		deallocate(xx)
		
		yGauss = yGauss / maxval(yGauss)
		yStep = yStep / maxval(yStep)
		
	end subroutine 

!------------------------------------------------------------
! Confidence intervals
! INPUT
!   xdata : x position of the histogram
!   ydata : y position of the histogram
! OUTPUT
!   est : estimated value of the parameter
! errup : value of x at the upper confidence interval
! errdown : value of x at the lower confidence interval
! conf : desired confidence level
!------------------------------------------------------------
	subroutine conf_limits_ccf(xdata, ydata, conf, est, errup, errdown)
	real(kind=8) :: xdata(:), ydata(:)
	real(kind=8) :: conf, est, errup, errdown
	real(kind=8) :: xpdf_low, xpdf_up, xpdf_mid, xmin, xmax, lower, upper, norm
	real(kind=8), allocatable :: xx(:), yy(:), x(:), y(:), xF(:), F(:)
	integer, allocatable :: ind(:)
	integer :: n, npto, loc(1), i
	
		n = size(xdata)
		
		xpdf_low = 0.50 - conf/200.
		xpdf_mid = 0.5
		xpdf_up  = 0.5 + conf/200.
		
		if (xpdf_low < 0.0 .or. xpdf_up > 1) then 
			print *,'wrong value for CONF'
			stop
		endif
				
		allocate(xx(n))
		allocate(yy(n))
		allocate(ind(n))
		
		xx = xdata
		yy = ydata / maxval(ydata)
						
! Sorting
		call qsortd(xx,ind,n)
		xx = xx(ind)
		yy = yy(ind)
		deallocate(ind)
						
! Interpolation by splines
		npto = 300
		xmin = xx(1)
		xmax = xx(n)
		
		if (npto > n) then
			allocate(x(npto))
			allocate(y(npto))
			
			do i = 1, npto
				x(i) = xmin + (xmax-xmin)/npto * i
			enddo
			call spline(xx,yy,x,y)
			n = npto
		else
			allocate(x(n))
			allocate(y(n))
			x = xx
			y = yy
		endif
		
! Peak normalization
		y = y / maxval(y)
		
! Computing the cumulative distribution function
		norm = int_tabulated(x,y)
		
		allocate(xF(n-1))
		allocate(F(n-1))
		xF = x(1:n)
				
		do i = 5, n
			F(i-1) = int_tabulated(x(1:i-1),y(1:i-1)) / norm
		enddo
				
		loc = minloc(abs(F - xpdf_low))		
		lower = xF(loc(1))
		
		loc = minloc(abs(F - xpdf_mid))		
		est = xF(loc(1))
		
		loc = minloc(abs(F - xpdf_up))
		upper = xF(loc(1))
		
		errup = upper - est
		errdown = est - lower

		deallocate(x,y)
		deallocate(xF,F)
		deallocate(xx,yy)
	
	end subroutine conf_limits_ccf

!------------------------------------------------------------
! Analyze the chains
!------------------------------------------------------------
	subroutine analyze_chains_multinest(chain_analysis, prior)
	type(markov_chain_type) :: chain_analysis	
	type(prior_type) :: prior	
	real(kind=8), pointer :: x(:), yGauss(:), yStep(:)
	integer :: i, j, k, n, loop, l
	character(len=120) :: str_parameter
	real(kind=8) :: est, errup1s, errup2s, errdown1s, errdown2s, chi2_mean, chi2_max, percentiles(7)
	integer, pointer :: indx(:)
	real(kind=8), allocatable :: chain_function(:)
	
		call read_chains_multinest(chain_analysis%filename, chain_analysis)
		
		open(unit=12,file=trim(adjustl(chain_analysis%filename))//'.hist1D',&
			action='write',status='replace')
		write(12,*) chain_analysis%nparam
		
		open(unit=13,file=trim(adjustl(chain_analysis%filename))//'.confidence',&
			action='write',status='replace')

		open(unit=14,file=trim(adjustl(chain_analysis%filename))//'.percentiles',&
			action='write',status='replace')
		write(14,*) 'Percentiles'
		write(14,*) '     0.2%       2.5%        16%         50%         84%        97.5%      99.8%'

! Do histograms for all inverted variables
		loop = 1
! 		do i = 1, chain_analysis%nparam


		do j = 1, n_param_per_component
			do i = 1, atmos%ncomponents
				do k = 1, atmos%n_nodes(i,j)
					
					if (atmos%priors(i)%typ(j) /= 2) then
						call oned_histogram(chain_analysis%chain(loop,:), x, yGauss, yStep)

! 1sigma confidence intervals
						call conf_limits_ccf(x, yStep, one_sigma, est, errup1s, errdown1s)
						call conf_limits_ccf(x, yStep, two_sigma, est, errup2s, errdown2s)
					else				
						allocate(x(10))
						x = atmos%priors(i)%mu(j)
						allocate(yGauss(10))
						yGauss = 1.0
						allocate(yStep(10))
						yStep = 1.0
						est = atmos%priors(i)%mu(j)
						errup1s = 0.d0
						errup2s = 0.d0
						errdown1s = 0.d0
						errdown2s = 0.d0
					endif				
					
					chain_analysis%most_probable(loop) = est
					write(13,*) est, chain_analysis%best_parameters(loop), errdown1s, errup1s, errdown2s, errup2s
 					str_parameter = atmos%priors(i)%param_names(j)
 					if (k == 1) write(*,*) '=============================================================================='
					write(*,FMT='(A,A,F12.4,A7,F12.4)') trim(adjustl(str_parameter)), ' : E(x)=', est, ' - MAP=', &
						chain_analysis%best_parameters(loop)
					write(*,FMT='(A,F12.4,A,F12.4)') '1-sigma : -', errdown1s, '  +', errup1s
					write(*,FMT='(A,F12.4,A,F12.4)') '2-sigma : -', errdown2s, '  +', errup2s
					n = size(x)
					write(12,*) loop, n
					do l = 1, n
						write(12,*) x(l), yGauss(l), yStep(l)
					enddo
					deallocate(x)
					deallocate(yGauss)
					deallocate(yStep)

! Calculate the percentiles
					percentiles = percentile_summary(chain_analysis%chain(loop,:))
					write(14,FMT='(7(F12.4))') percentiles
					
					loop = loop + 1
				enddo
			enddo
		enddo


! Calculate average ln L for calculating Kullback-Leibler distance
		chain_analysis%avg_lnL = sum(chain_analysis%chain(chain_analysis%nparam+1,:)) / chain_analysis%nlength
		
		write(13,FMT='(5(A,F12.5))') 'Z=', chain_analysis%evidence, ',  <ln L>=', chain_analysis%avg_lnL, &
			',  D_KL=', -chain_analysis%evidence + chain_analysis%avg_lnL, ', logL_best=', chain_analysis%best_likelihood,&
			', chi2=', -2.d0 * chain_analysis%best_likelihood / Params%n_data

		write(*,FMT='(5(A,F12.5))') 'Z=', chain_analysis%evidence, ',  <ln L>=', chain_analysis%avg_lnL, &
			',  D_KL=', -chain_analysis%evidence + chain_analysis%avg_lnL, ', logL_best=', chain_analysis%best_likelihood,&
			', chi2=', -2.d0 * chain_analysis%best_likelihood / Params%n_data
			
		close(12)
		close(13)

! Write a file with the number of nodes and their locations
		open(unit=12,file=trim(adjustl(chain_analysis%filename))//'.nodes',&
			action='write',status='replace')
		write(12,*) 'Number of parameters per component / number of components'
		write(12,*) n_param_per_component, atmos%ncomponents
		write(12,*) 'Number of nodes per parameter and location of nodes'
		do j = 1, n_param_per_component
			do i = 1, atmos%ncomponents
				str_parameter = atmos%priors(i)%param_names(j)
				write(12,*) trim(adjustl(str_parameter))
				write(12,*) atmos%n_nodes(i,j)
				write(12,*) (atmos%pos_nodes(i,j,k),k=1,atmos%n_nodes(i,j))
			enddo
		enddo		
		close(12)

! Write a file with the most probable model
! The call to the synthesis puts the atmosphere in hydrostatic equilibrium

! Synthesize the MAP profile
		call synthesize_profiles(chain_analysis%most_probable, obs%model_eval)
		
		do i = 1, atmos%ncomponents

			write(str_parameter,FMT='(I1)') i
			open(unit=12,file=trim(adjustl(chain_analysis%filename))//'.MAP'//trim(adjustl(str_parameter)),&
				action='write',status='replace')
			write(12,*) 'Format version: 1.0'
			write(12,*) '# tau5000 T      P_el       v_mic        B        v_los   Inclin  Azimuth'
			write(12,*) Params%n_points
			write(12,*) atmos%perturbed(i)%v_mac, atmos%perturbed(i)%stray, atmos%ff(i)
			do j = 1, Params%n_points
				write(12,FMT='(F5.2,1X,F7.1,4(1X,ES10.2),1X,F6.1,1X,F6.1)') &
					atmos%perturbed(i)%ltau_500(j), atmos%perturbed(i)%temp(j), atmos%perturbed(i)%el_p(j), &
					atmos%perturbed(i)%v_mic(j), atmos%perturbed(i)%b_str(j), atmos%perturbed(i)%v_los(j), atmos%perturbed(i)%b_inc(j), &
					atmos%perturbed(i)%b_azi(j)
			enddo
			close(12)

			open(unit=12,file=trim(adjustl(chain_analysis%filename))//'.original'//trim(adjustl(str_parameter)),&
				action='write',status='replace')
			write(12,*) 'Format version: 1.0'
			write(12,*) '# tau5000 T      P_el       v_mic        B        v_los   Inclin  Azimuth'
			write(12,*) Params%n_points
			write(12,*) atmos%stratification(i)%v_mac, atmos%stratification(i)%stray, atmos%ff(i)
			do j = 1, Params%n_points
				write(12,FMT='(F5.2,1X,F7.1,4(1X,ES10.2),1X,F6.1,1X,F6.1)') &
					atmos%stratification(i)%ltau_500(j), atmos%stratification(i)%temp(j), atmos%stratification(i)%el_p(j), &
					atmos%stratification(i)%v_mic(j), atmos%stratification(i)%b_str(j), atmos%stratification(i)%v_los(j), atmos%stratification(i)%b_inc(j), &
					atmos%stratification(i)%b_azi(j)
			enddo
			close(12)
			
		enddo
		

! Save MAP profile
		open(unit=12,file=trim(adjustl(chain_analysis%filename))//'.synthesis',&
			action='write',status='replace')
		do i = 1, obs%npoints
			write(12,FMT='(F10.4,2X,4(E12.4))') obs%obs_x(i), obs%model_eval(1,i), obs%model_eval(2,i), &
 				obs%model_eval(3,i), obs%model_eval(4,i)
 		enddo
 		close(12)
		
	end subroutine analyze_chains_multinest

end module chain_analysis_mod