! This subroutine computes the damping parameter for the absorption profile.
! Temp, El_P and Pg and in cgs units
! If GA and GQ are .gt. 0, they are used to compute the Stark and
! radiative broadenings in the same manner as MULTI.
! If GA and GQ are .le. 0 then an approximation is used (see Gray)
!
Subroutine damping(Line, Temp, El_p, Pg, Dldop, Damp, GA, GQ)
  Use Param_structure
  Use Atomic_data
  Use Line_data_structure
  Implicit None
  Type (Line_data) :: Line
  Real :: Temp, El_p, Dldop, Damp, C6, ioniz, chi_l, a, b
  Real :: gamma_r, gamma_vdw, gamma_s, Pg, dldopHz, GA, GQ
  Real :: Sigma, Alpha, X, GX, GAMMAF, GVW, K, A0, M0, H1FRC, HE1FRC, VBAR
  Real, Dimension (10) :: Pp
  Logical, Save :: Warning
  Data Warning/.FALSE./
  PARAMETER (K=1.380658E-23,M0=1.660540E-27)
  PARAMETER (A0=5.29177249E-11)
!
  dldopHz=cc/Line%Wlength/Line%Wlength*Dldop/1.E-8 ! Dldop in Hz
!
! Radiative damping:
  If (GA .le. 0) then 
! Reference: Gray 1976, "The observation and analysis of stellar 
!  photospheres", pag 227, just after Eq. (11-19). This approximation
!  is poor, but the radiative damping is usually negligible anyway.
!
     gamma_r=0.22233/(Line%Wlength*Line%Wlength*1.E-16)
  Else
     gamma_r=GA ! Same as MULTI
  End if
!
! Van der Waals damping:
  If (Line%collisions .eq. 1) then 
     ! Reference: Gray (see above, pag 239, Eqs. (11-35) and (11-36))
     ! Formula used: 
     !    log (gamma_vdw) = 19.6 + 2./5.*log (C6(H)) + log Pg - 7./10.*log T
     If (Line%Ion_stage .eq. 1) then
        ioniz=At_ioniz1(Line%Atomic_number)
     else if (Line%Ion_stage .eq. 2) then
        ioniz=At_ioniz2(Line%Atomic_number)
     else
        If (.not. Warning) &
             Print *,'Ionization stage gt 2. Van der Waals broadening not considered'
        ioniz=-1
        Warning=.TRUE.
     end if
     gamma_vdw=0.
     If (ioniz .gt. -1) then 
        chi_l=1.24E4/Line%Wlength
        a=(ioniz-Line%Energy_low-chi_l)**2.
        b=(ioniz-Line%Energy_low)**2.
        C6=3.E-31*(1./a-1./b)
        gamma_vdw=19.6 + 2./5.*log10(C6) + log10(Pg) - 7./10.*log10(Temp)
        gamma_vdw=10.**(gamma_vdw)
        gamma_vdw=gamma_vdw*Line%VDW_enh ! Empirical Van der Waals enhancement
        If (C6 .lt. 0) then
           gamma_vdw=0.
        End if
     End if
  Else if (Line%collisions .eq. 2) then ! Barklem formula
                           ! Following
                           ! http://www.astro.uu.se/~barklem/howto.html
     Sigma=Line%Bark_sigma*A0*A0
     Alpha=Line%Bark_alpha
    !  Compute the Gamma function of X, this function is valid over the 
    !  range 1<X<2 ie. 0<ALPHA<2 which is always satisfied    
     X=2.-ALPHA*.5
     GX=X-1.0
     GAMMAF=1+(-.5748646+(.9512363+(-.6998588+(.4245549-.1010678*GX &
          )*GX)*GX)*GX)*GX
    ! Compute the halfwidth per unit perturber number density for this temp
     GVW=(4./PI)**(ALPHA*0.5)*GAMMAF*1.E4*SIGMA
     VBAR=SQRT(8.*K*Temp/PI/M0*(1./1.008+1./Line%Atomic_weight))
     GVW=GVW*((VBAR/1.E4)**(1.-ALPHA))
    ! Get H and He partial pressures
     Call Gasc(Temp, El_P, Pg, Pp)
    ! Fullwidth given H1FRC perturbers per cm^3 with approx He I broadening
    ! The factor of 2 comes from converting to the full width.
    ! The factor of 1.E6 converts from SI to cgs
     H1FRC=Pp(1)*(Pg/BK/Temp) ! Assuming a H gas (particles per cm^3)
     HE1FRC=Pp(2)*(Pg/BK/Temp) ! Assuming a H gas (particles per cm^3)
     GVW=GVW*(H1FRC+ 0.42*HE1FRC)*1.E6*2.
     gamma_vdw=GVW
     gamma_vdw=gamma_vdw*Line%VDW_enh ! Empirical Van der Waals enhancement
  Else
     Print *,'Unknown collisional broadening in damping.f90'
     Stop
  Endif
!
! Stark damping
  If (GQ .le. 0) then
! Formula used: gamma_4=38.8*(v**(1./3.))*(C4**(2./3.))*N , from
!   Unsold 1955, "Physik der Sternatmospharen", 2nd ed., 
!   Springer-Verlag, pp. 326 and 331. According to Gray (ref above), 
!   pag 237, this is similar to 
!   log (gamma_4) = 19.4 + 2./3*log C4 + log Pe - 5./6.*log T
!   The value of log C4 is approximated by -14. (see Gray, Table 11-3)
     gamma_s = 19.4 + 2./3.*(-14.) + log10(El_p) - 5./6.*log10(Temp)
     gamma_s = 10.**(gamma_s)
  Else
     gamma_s = GQ*El_p/bk/Temp ! Same as MULTI
  End if
  Damp=(gamma_r+gamma_vdw+gamma_s)/4./Pi/dldopHz

  Return
End Subroutine damping
