! This subroutine solves the forward problem, i.e., the synthesis of the
! spectral lines assuming LTE.
!
! IMPORTANT NOTE: This routine puts the model in hydrostatic equilibrium to
! re-evaluate the electron pressure, gas pressure, density and height scale
! (although only the electron pressure is actually used in this routine).
! If you do not wish to impose hydrostatic equilibrium, comment out the
! call to HYDROSTATIC at the beginning of this routine. In that case
! the inversion procedure would not touch the electron pressure.
!
! v3.0: * Uses a Neural Network to convert between P_g and P_el (if abundances
!        are the default Greveese ones)
!
!       * Changed the loop structure so that PROFILES is called once for each
!        depth point. All wavelengths are computed in the same call.
!
! v3.1: * Several optimization improvements, including search for grid
!        in which to compute profiles
!       * Introduced gravity as input parameter
!       * Mu=0. means compute stellar flux using Gaussian quadrature. By 
!        default, 3 points are used. This can be changed in the definitions
!        of NQUAD below (between 2 and 5!!!).
!
! v3.2: * Added Error to argument list to inform caller when something 
!        goes wrong
!
! navarro@hao.ucar.edu, Mar 2000
!
! Version control:
!
Module Forward_version
  Character (len=25) :: Forward_ver='NICOLE Forward v3.2'
End Module Forward_version
!
Subroutine Forward(Params, Line, Region, Atmo, Syn_profile, Hydro, Error)
  Use Param_structure
  Use Model_structure
  Use Model_operations
  Use NLTE_Atom_structure
  use NLTE_inp
  Use NLTE_vars
  Use Atomic_data
  Use Gauss_quad
  Use Line_data_structure
  Use Background_opacity_module, Only: Background_opacity
  Implicit None
  Type (Parameters) :: Params
  Type (Line_data), dimension (Params%n_lines) :: Line
  Type (Region_data), dimension (Params%n_regions) :: Region
  Type (NLTE_Atom), Save :: Atom
  Type (NLTE_input), Save :: NLTEInput
  Type (NLTE_variables), Save :: NLTE
  Type (Model) :: Atmo
  Integer, Parameter :: NQuad=3 ! Number of points in quadrature (2 to 5)
  Logical :: Hydro, Incomplete, End
  Logical :: Error
  Real, Dimension (Params%n_data) :: Syn_profile
  Real, Dimension (Params%n_points) :: Source_f, ltau_500_mu
  Real, Dimension (Params%n_points) :: Cont_op_5000, Cont_op
  Real, Dimension (Params%n_points) :: ng_i, ng_j
  Real, Dimension (Params%n_lines, Params%n_points) :: Dldop, Damp, Line_op
  Real, Dimension (4, 4, Params%n_points) :: Absorp_height
  Real, Dimension (:,:,:,:), Allocatable :: Absorp, TotAbsorp
  Real, Dimension (:,:), Allocatable :: Phi_I, Phi_Q, Phi_U, Phi_V
  Real, Dimension (:,:), Allocatable :: Psi_Q, Psi_U, Psi_V, TotEmis
  Real, Dimension (:), Allocatable :: Wave
  Integer, Dimension (:), Allocatable :: IndexWave
  Real, Dimension (10) :: Pp
  Real, Dimension (4) :: Stokes
  Real, Dimension (3) :: x, y
  Real, Dimension (1) :: imin1
  Real, Dimension (NQuad) :: XMU, WMU
  Real :: nu, Wave_cm, Av_mol_weight
  Real :: conhsra, syn_hsra, reference_cont, mu, DWave, last_update
  Real :: PH, PHminus, PHplus, PH2, PH2plus, n2P, Scat
  Integer :: npoints, idepth, iline, iwave, nwlengths, idata, ichoice, iunit, ntrans
  Integer :: iregion, i, j, nw, istart, iend, NMu, imu, itran, imin
  Integer, Parameter :: nformalsolutions=2
  Integer, Dimension(nformalsolutions) :: nformal
  Logical, Save :: First_time=.TRUE., do_NLTE
  Logical :: CheckNaN
  Character (len=256) :: String

  If (First_time) then
     Call Read_abundances
     ! Check if abundances are default (check only first 30 elements)
     Params%def_abund=1
     Do i=1, 30
        if (Abs(At_abund(i)-At_Greveese_abund(i)) .gt. 0.01) Params%def_abund=0
     End do
     If (Params%speed .lt. 2) &
          Params%def_abund=0 ! Force to not use ANN

     If (Params%def_abund .ge. 1 .and. Params%printout .gt. 0) then
        Print *,'Abundances are known. Using Neural Network to compute gas pressure'
     End if
     ! Need to do NLTE calculation or are we in pure LTE mode?
     do_NLTE=.False.
     Call Open_file(iunit, 'NLTE_lines')
     End=.False.
     ntrans=0.
     Do While (.not. End)
        Call Read_next_nocomment(iunit, String, End)
        ntrans=ntrans+1
     End do
     ntrans=ntrans-1
     Close (iunit)
     If (ntrans .ge. 1) do_NLTE=.True.
     !
     If (do_NLTE) then
        Call NLTE_init(Params, NLTEinput, NLTE, Atom, Atmo)
        Call Read_NLTE_lines(Params, NLTEInput, NLTE, Atom, Line)
     End if
  End if
  First_time=.FALSE.
  Error=.False.
! If flux computation is required, set Gaussian weights
  If (Params%heliocentric .ne. 0) then
     NMu=1
     XMU(1)=Params%heliocentric
     WMU(1)=1.
  Else
     If (NQuad .gt. 5) then
        Print *,'Too many quadrature points'
        Stop
     End if
     NMu=NQuad
     XMU(1:NQuad)=XMU_Gauss_Quad(NQuad,1:Nquad)
     WMU(1:NQuad)=WMU_Gauss_Quad(NQuad,1:Nquad)
  End if
! Now let's put the model in hydrostatic equilibrium and fill in
! the columns Gas_P, El_p, Rho and Z_scale of the model.
!
  npoints=Params%n_points
  
 ! Transform velocities from km/s to cm/s
 	Atmo%v_mic(1:npoints) = Atmo%v_mic(1:npoints) * 1.e5
 	Atmo%v_los(1:npoints) = Atmo%v_los(1:npoints) * 1.e5


  If (Hydro) then
     Call Hydrostatic(Params, Atmo, Error)
  Else ! Trust El_P, rho and Gas_P and fill the nH, nHplus, nHminus,
       !       nH2 columns of the model
     Do idepth=1, npoints
        Call Gasc(Atmo%Temp(idepth), Atmo%El_p(idepth), Atmo%Gas_p(idepth), Pp)
        Atmo%nH(idepth)=Pp(1)/Pp(8)*Atmo%El_p(idepth)/bk/Atmo%Temp(idepth)
        Atmo%nHplus(idepth)=Pp(6)/Pp(8)*Atmo%El_p(idepth)/bk/Atmo%Temp(idepth)
        Atmo%nHminus(idepth)=Pp(10)/Pp(8)*Atmo%El_p(idepth)/bk/Atmo%Temp(idepth)
        Atmo%nH2(idepth)=Pp(7)/Pp(8)*Atmo%El_p(idepth)/bk/Atmo%Temp(idepth)
     End do
  End if
!
  nformal(1:nformalsolutions)=0. ! Initialize the f. s. counter
  idata=1 ! Index for Syn_profile
  Syn_profile(:)=-1e10
!
! Calculate line- and wavelength-independent data.
!
  Do idepth=1, npoints !$$ PARALLEL LOOP START
     n2P=BK*Atmo%Temp(idepth)
     Cont_op_5000(idepth)=Background_opacity(Atmo%Temp(idepth), Atmo%El_p(idepth), Atmo%nH(idepth)*n2P, &
          Atmo%nHminus(idepth)*n2P, Atmo%nHplus(idepth)*n2P, Atmo%nH2(idepth)*n2P, 0., 5000., Scat)
  End do !$$ PARALLEL LOOP ENDe
  Cont_op_5000=Cont_op_5000/Atmo%Rho ! Convert to cm^2/g
!
! Loop in spectral lines. Compute line-dependent, wave-independent data.
!
  Do iline=1,Params%n_lines !$$ PARALLEL LOOP START
     Dldop(iline, 1:npoints)=Line(iline)%Wlength/cc* & ! Doppler width in A
          Sqrt(2.*bk*Atmo%Temp(1:npoints)/&
          At_weight(Line(iline)%Atomic_number)/mass_pr+ &
          Atmo%v_mic(1:npoints)**2.)
     If (Line(iline)%NLTEtransition .eq. 0) then ! LTE line
        Do idepth=1, npoints
           Call Damping(Line(iline), Atmo%Temp(idepth), Atmo%El_p(idepth), &
                Atmo%Gas_p(idepth), Dldop(iline, idepth), &
                Damp(iline, idepth), -1.0, -1.0) ! Compute damping            
        End do
        Call LTE_pop(Params, Line(iline), Atmo, ng_i, ng_j)
     Else ! NLTE
        itran=Line(iline)%NLTEtransition
        Do idepth=1, npoints
           Call Damping(Line(iline), Atmo%Temp(idepth), Atmo%El_p(idepth), &
                Atmo%Gas_p(idepth), Dldop(iline, idepth), &
                Damp(iline, idepth), Atom%GA(itran), Atom%GQ(itran))
        End do
        NLTE%Atmo=Atmo
        If (NLTEInput%VelFree) NLTE%Atmo%v_los=0.
        Call NLTE_init(Params, NLTEinput, NLTE, Atom, Atmo)
        Call SolveStat(NLTE, NLTEInput, Atom)
        Do idepth=1, npoints
           If (CheckNaN(NLTE%N(1, idepth))) then
              Print *,'Error!! NLTE calculation returned NaN for populations at i=',idepth
              NLTE%Error=.True.
           End if
        End do
        i=Atom%i(itran)
        j=Atom%j(itran)
        ng_i(:)=NLTE%N(i,:)/Atom%g(i)*Line(iline)%NLTE_nl_ratio
        ng_j(:)=NLTE%N(j,:)/Atom%g(j)*Line(iline)%NLTE_nu_ratio
        Line(iline)%NLTESource_f(1:npoints,1:Line(iline)%NLTEgridsize)= &
             NLTE%Source_f(1:npoints,1:Atom%NQ(itran),itran)
     End if
     Line_op(iline, 1:npoints)=4.9947E-21*(10.**Line(iline)%loggf)* &
          (ng_i(1:npoints)-ng_j(1:npoints))/Dldop(iline, 1:npoints)* &
          (Line(iline)%Wlength**2) ! Still needs the profile
     Line_op(iline, 1:npoints)=Line_op(iline, 1:npoints) / &
          Atmo%Rho(1:npoints) ! cm^2/cm^3 to cm^2/g
  End do ! End line loop
!
! Start loop in spectral regions and wavelengths.
!
  Do iregion=1, Params%n_regions
     last_update=-10 ! Make sure that opacities will be computed the first time
     nwlengths=Region(iregion)%nwavelengths
     If (Params%Reference_cont .eq. 0) then ! Do not normalize
        reference_cont=1.e15 ! For numerical precision. Will remove this factor
                             ! at the end
     Else If (Params%Reference_cont .eq. 1) then
        reference_cont=conhsra(Region(iregion)%First_wlength + &
             Region(iregion)%nwavelengths * &
             Region(iregion)%Wave_step/2.) ! HSRA continuum at middlepoint
     Else If (Params%Reference_cont .eq. 2) then
        reference_cont=conhsra(5000.) ! HSRA continuum at 500nm
     Else If (Params%Reference_cont .eq. 3) then
      reference_cont=syn_hsra(Region(iregion)%First_wlength + &
             Region(iregion)%nwavelengths * &
             Region(iregion)%Wave_step/2., Params%heliocentric)! HSRA continuum
     Else
        Print *,'Unknown spectrum normalization option'
        Stop
     End if

! Compute wavelength profiles
     If (Allocated(TotAbsorp)) then
        Deallocate (TotAbsorp)
        Deallocate (TotEmis)
        Deallocate (Absorp)
        Deallocate (Phi_I)
        Deallocate (Phi_Q)
        Deallocate (Phi_U)
        Deallocate (Phi_V)
        Deallocate (Psi_Q)
        Deallocate (Psi_U)
        Deallocate (Psi_V)
        Deallocate (Wave)
     End if
     Allocate (TotAbsorp(4, 4, nwlengths, Params%n_points))
     Allocate (TotEmis(nwlengths, Params%n_points))
     Allocate (Absorp(4, 4, nwlengths, Params%n_points))
     Allocate (Phi_I(nwlengths, Params%n_points))
     Allocate (Phi_Q(nwlengths, Params%n_points))
     Allocate (Phi_U(nwlengths, Params%n_points))
     Allocate (Phi_V(nwlengths, Params%n_points))
     Allocate (Psi_Q(nwlengths, Params%n_points))
     Allocate (Psi_U(nwlengths, Params%n_points))
     Allocate (Psi_V(nwlengths, Params%n_points))
     Allocate (Wave(nwlengths))
     TotAbsorp(1:4, 1:4, 1:nwlengths, 1:Params%n_points)=0.
     TotEmis(1:nwlengths, 1:Params%n_points)=0.
     Do iwave=1, nwlengths
        Wave(iwave)=Region(iregion)%First_wlength + &
             (iwave-1)*Region(iregion)%Wave_step ! Lambda in A
     End do
     Do iline=1, Params%n_lines ! Loop in lines to blend
        If (MaxVal(Line_op(iline,:)/Cont_op_5000(:)) .gt. 1.e-1) then ! Consider line
           Incomplete=.TRUE.
           Do While (Incomplete)
              istart= (Line(iline)%Wlength-Line(iline)%Width - Wave(1))/ &
                   Region(iregion)%Wave_step -1
              iend= (Line(iline)%Wlength+Line(iline)%Width - Wave(1))/ &
                   Region(iregion)%Wave_step +1
              If (istart .gt. nwlengths) istart=nwlengths-2
              If (istart .lt. 1) istart=1
              If (iend .lt. 1) iend=3
              If (iend .gt. nwlengths) iend=nwlengths
              If (iend .le. istart+2) then
                 istart=istart-2
                 iend=istart+2
              End if
              If (istart .lt. 1) istart=1
              If (iend .gt. nwlengths) iend=nwlengths
              nw=0
              Do iwave=istart, iend, Params%Skip_lambda
                 nw=nw+1
              End do
              If (Allocated(IndexWave)) Deallocate(IndexWave)
              Allocate (IndexWave(nw))
              nw=0
              Do iwave=istart, iend, Params%Skip_lambda
                 nw=nw+1
                 IndexWave(nw)=iwave
              End do
              Call Profiles(Params, Line(iline), nwlengths, nw, IndexWave, Wave(:), Atmo, &
                   Damp(iline,:), Dldop(iline,:), Line_op(iline,:), &
                   Phi_I(:,:), Phi_Q(:,:), &
                   Phi_U(:,:), Phi_V(:,:), Psi_Q(:,:), &
                   Psi_U(:,:), Psi_V(:,:), &
                   Cont_op_5000(:), Incomplete) ! Compute profiles
              If (istart .eq. 1 .and. iend .eq. nwlengths) Incomplete=.FALSE.
              If (Line(iline)%Width .gt. 5.) Incomplete=.FALSE.
              If (Incomplete) Line(iline)%Width=Line(iline)%Width*2.
           End do ! While incomplete
           Call Abs_matrix(npoints, nwlengths,  Phi_I(:,:), Phi_Q(:,:), &
                Phi_U(:,:), Phi_V(:,:), &
                Psi_Q(:,:), Psi_U(:,:), Psi_V(:,:), &
                Cont_op_5000, Absorp(:,:,:,:)) ! Build normalized absorp. m.
           Do iwave=1, nwlengths, Params%skip_lambda
              DWave=Wave(iwave)-Line(iline)%Wlength
              If (Line(iline)%NLTEtransition .ge. 1) then
                 If (Dwave .lt. MinVal(Line(iline)%NLTEgrid)) DWave=Abs(DWave)
              Endif
              nu=cc/(Wave(iwave)*1e-8) ! s^-1
              Wave_cm=cc/nu
              If (Line(iline)%NLTEtransition .gt. 0) then
                 If (DWave .ge. MinVal(Line(iline)%NLTEgrid) .and. &
                   DWave .le. MaxVal(Line(iline)%NLTEgrid)) then ! Interpolate NLTE source function
                    Do idepth=1, Params%n_points
                       Call parab_interp(Line(iline)%NLTEgridsize,&
                            Line(iline)%NLTEgrid(:), &
                            Line(iline)%NLTESource_f(idepth,:),1, &
                            DWave,Source_f(idepth))
                    End do
! S has units of 2*hh*nu^3/cc^2, we need  2*hh*c^2/(lambda^5).
                    Source_f(:)=Source_f(:)*(nu/cc)*nu
                    TotEmis(iwave,:)=TotEmis(iwave,:)+ &
                         Absorp(1,1,iwave,:)*Source_f(:)* &
                         Line(iline)%b_up(:) ! Apply departure coefficients
                 Else ! LTE
                    TotEmis(iwave,:)=TotEmis(iwave,:)+Absorp(1,1,iwave,:)* &
                         2.*hh/Wave_cm*cc*cc/(Wave_cm**4)/  &
                         (exp(hh*nu/bk/Atmo%Temp(:))-1.)* &   ! Planck f (cgs)
                         Line(iline)%b_up(:) ! Apply departure coefficients
                 End if
              Else ! LTE
                 TotEmis(iwave,:)=TotEmis(iwave,:)+Absorp(1,1,iwave,:)* &
                      2.*hh/Wave_cm*cc*cc/(Wave_cm**4)/  &
                      (exp(hh*nu/bk/Atmo%Temp(:))-1.)* &   ! Planck f (cgs)
                      Line(iline)%b_up(:) ! Apply departure coefficients
              End if
           End do ! Do in iwave
           Do idepth=1, Params%n_points ! Apply depature coefficients
              Absorp(:,:,:,idepth)=Absorp(:,:,:,idepth)* &
                   Line(iline)%b_low(idepth)
           End do
           TotAbsorp(:,:,:,:)=TotAbsorp(:,:,:,:)+ &
                Absorp(:,:,:,:) ! Total absorption of all blends (normalized)
        End if ! Consider line
     End do ! End loop in lines to blend
     Do iwave=1, nwlengths, Params%skip_lambda !$$ PARALLEL LOOP START
        If (abs(last_update-Wave(iwave)) .gt. Params%Update_opac) then !Update background opacities
           nu=cc/(Wave(iwave)*1e-8) ! s^-1
           Wave_cm=cc/nu
           Do idepth=1, npoints
              n2P=BK*Atmo%Temp(idepth)
              Cont_op(idepth)=Background_opacity(Atmo%Temp(idepth), Atmo%El_p(idepth), Atmo%nH(idepth)*n2P, &
                   Atmo%nHminus(idepth)*n2P, Atmo%nHplus(idepth)*n2P, Atmo%nH2(idepth)*n2P, 0., Wave(iwave), Scat)
           End do
           Cont_op=Cont_op/Atmo%Rho ! Convert to cm^2/g
           last_update=Wave(iwave)
        End if
 ! Add continuum opacity and emission at this wlength
        Do i=1, 4
           TotAbsorp(i, i, iwave, 1:Params%n_points)= &
                TotAbsorp(i, i, iwave, 1:Params%n_points)+ &
                Cont_op(1:Params%n_points)/ &
                Cont_op_5000(1:Params%n_points)
        End do
        TotEmis(iwave,:)=TotEmis(iwave,:)+Cont_op(:)/Cont_op_5000(:)* &
              2.*hh/Wave_cm*cc*cc/(Wave_cm**4)/  &
                      (exp(hh*nu/bk/Atmo%Temp(1:npoints))-1.) ! Planck f (cgs)

        Source_f=TotEmis(iwave,:)/TotAbsorp(1,1,iwave,:)
!
        Syn_profile(idata:idata+3)=0.
        Absorp_height(:,:,:)=TotAbsorp(:,:,iwave,:)
        Do imu=1, NMu ! Loop in angles to compute flux
           mu=XMU(imu) ! mu=Cos. of heliocentric angle
           ltau_500_mu=Atmo%ltau_500 - Log10(mu) ! Optical depth along l.o.s.
           Call formal_solution(Params%n_points, Params%formal_solution, &
                ltau_500_mu, Absorp_height, Source_f, &
                Stokes, ichoice) ! Formal solution
           Syn_profile(idata:idata+3)=Syn_profile(idata:idata+3) + &
                WMu(imu)*Stokes(1:4)/reference_cont ! Store result
        End do
        idata=idata+4*Params%Skip_lambda ! Update the data index
        If (ichoice .gt. nformalsolutions) then
           Print *,'Error in formal solution. Unknown formal solution method'
           Stop
        End if
        nformal(ichoice)=nformal(ichoice)+1 ! How many solutions of each type
     End do !$$ PARALLEL LOOP END
     Deallocate (TotAbsorp)
     Deallocate (TotEmis)
     Deallocate (Absorp)
     Deallocate (Phi_I)
     Deallocate (Phi_Q)
     Deallocate (Phi_U)
     Deallocate (Phi_V)
     Deallocate (Psi_Q)
     Deallocate (Psi_U)
     Deallocate (Psi_V)
     Deallocate (Wave)
     If (Allocated(IndexWave)) Deallocate (IndexWave)
  End do !$$ PARALLEL LOOP END
  If (Params%printout .ge. 3) then
     Print *,'Number of Stokes formal solutions using Hermite: ',nformal(1)
     Print *,'Number of Stokes formal solutions using WPM: ',nformal(2)
  End if
  If (Params%Reference_cont .eq. 0) & ! Remove normalization factor
       Syn_profile=Syn_profile*1.e15

  If (Params%Skip_lambda .eq. 1) &
       Call Convolve_profile(Params, Region, Atmo, Syn_profile) ! Macroturbulence
  Call Add_stray_light(Params, Atmo, Region, Syn_profile) ! Add stray light &
!
  If (NLTE%Error) Error=.True.
  Return
!
  End Subroutine Forward
